/*
 * Velbus Serial Communications Library for .NET
 * Copyright (C) 2007 Velleman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 */
#region "Imports"

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace Velleman.Velbus
{
    /// <summary>
    /// Packet priority. This is important for the Velbus network only,
    /// the dispatcher does not yet keep track of priorities.
    /// </summary>
    public enum PacketPriority
    {
        /// <summary>
        /// High priority. Usually used by real-time command like
        /// pressing a button etc.
        /// </summary>
        High = 0xF8,

        /// <summary>
        /// Low priority.
        /// </summary>
        Low = 0xFB
    }

    /// <summary>
    /// Velbus protocol packet.
    /// </summary>
    public class Packet : ICloneable
    {
        #region "Public Fields"

        /// <summary>
        /// Maximum packet size used by the Velbus protocol.
        /// </summary>
        public const int MAX_PACKET_SIZE = 14;

        /// <summary>
        /// Start of packet.
        /// </summary>
        public const byte STX = 0x0F;

        /// <summary>
        /// End of packet.
        /// </summary>
        public const byte ETX = 0x04;

        #endregion

        /// <summary>
        /// Raw representation of this packet.
        /// </summary>
        private byte[] rawPacket = new byte[MAX_PACKET_SIZE];

        #region "Constructors"

        /// <summary>
        /// Initializes a new instance of the Packet class.
        /// </summary>
        public Packet()
            : this(0x00)
        {}

        /// <summary>
        /// Initializes a new instance of the Packet class using the
        /// specified address.
        /// </summary>
        /// <param name="address">Address to use.</param>
        public Packet(int address)
            : this(address, PacketPriority.High)
        {}

        /// <summary>
        /// Initializes a new instance of the Packet class using the
        /// specified address and priority.
        /// </summary>
        /// <param name="address">Address to use.</param>
        /// <param name="priority">Packet priority.</param>
        public Packet(int address, PacketPriority priority)
            : this(address, priority, 0)
        {}

        /// <summary>
        /// Initializes a new instance of the Packet class using the
        /// specified address, priority and data size.
        /// </summary>
        /// <param name="address">Address to use.</param>
        /// <param name="priority">Packet priority.</param>
        /// <param name="dataSize">Number of databytes.</param>
        public Packet(int address, PacketPriority priority, int dataSize)
            : this(address, priority, dataSize, false)
        {}

        /// <summary>
        /// Initializes a new instance of the Packet class using the
        /// specified address, priority, data size and rtr state.
        /// </summary>
        /// <param name="address">Address to use.</param>
        /// <param name="priority">Packet priority.</param>
        /// <param name="dataSize">Number of databytes.</param>
        /// <param name="rtr">Rtr state.</param>
        public Packet(int address, PacketPriority priority, int dataSize, bool rtr)
        {
            Address = address;
            Priority = priority;
            DataSize = dataSize;
            Rtr = rtr;
        }

        #endregion

        #region "Public Properties"

        /// <summary>
        /// Gets or sets the address of the packet.
        /// </summary>
        public int Address
        {
            get
            {
                return rawPacket[2];
            }
            set
            {
                if ((value < 0) || (value > 0xFF))
                {
                    throw new ArgumentOutOfRangeException("Address", Address,
                    "Valid address range between 0 and 255.");
                }
                rawPacket[2] = (byte)value;
            }
        }

        /// <summary>
        /// Gets or sets the datasize of the packet.
        /// </summary>
        public int DataSize
        {
            get
            {
                return (rawPacket[3] & 0x0F);
            }
            set
            {
                if ((value < 0) || (value > 8))
                {
                    throw new ArgumentOutOfRangeException("DataSize", DataSize,
                        "Invalid datasize. Value must range between 0 and 8.");
                }
                rawPacket[3] = (byte)((rawPacket[3] & 0xF0) + value);
            }
        }

        /// <summary>
        /// Gets the total size of the packet in its raw form.
        /// </summary>
        public int Size
        {
            get
            {
                return (DataSize + 6);
            }
        }

        /// <summary>
        /// Gets or sets the priority of the packet.
        /// </summary>
        public PacketPriority Priority
        {
            get
            {
                return (rawPacket[1] == 0xF8 ? PacketPriority.High : PacketPriority.Low);
            }
            set
            {
                rawPacket[1] = (byte)(value == PacketPriority.High ? 0xF8 : 0xFB);
            }
        }

        /// <summary>
        /// Gets or sets the request to reply state of the packet.
        /// </summary>
        public bool Rtr
        {
            set
            {
                if (value)
                    rawPacket[3] |= 0x40;
                else
                    rawPacket[3] &= 0x0F;
            }
            get
            {
                return ((rawPacket[3] & 0x40) == 0x40);
            }
        }

        /// <summary>
        /// Gets or sets the databyte at the specified position.
        /// </summary>
        /// <param name="idx">Index of the databyte to get or set.</param>
        /// <returns>Returns the requested databyte.</returns>
        public byte this[int idx]
        {
            get
            {
                return rawPacket[4 + idx];
            }
            set
            {
                rawPacket[4 + idx] = value;
            }
        }

        /// <summary>
        /// Gets or sets the command byte of the packet. Since the command
        /// byte is the first databyte, the datasize needs to be greater or 
        /// equal to one.
        /// </summary>
        public int Command
        {
            get
            {
                if (DataSize <= 0)
                    throw new ArgumentOutOfRangeException("Packet does not contain a command byte.");
                return this[0];
            }
            set
            {
                if (DataSize <= 0)
                    throw new ArgumentOutOfRangeException("Packet does not contain a command byte.");
                this[0] = (byte)value;
            }
        }

        /// <summary>
        /// Checks if the packet has a command byte (eg. if DataSize >= 1).
        /// </summary>
        /// <returns></returns>
        public bool HasCommand()
        {
            return (DataSize >= 1);
        }

        #endregion

        #region "Public Methods"

        /// <summary>
        /// Packs the byte so it is ready for sending. Packing involves
        /// adding a checksum and the frame delimiters.
        /// </summary>
        /// <returns>Returns a raw representation of the packet.</returns>
        public byte[] Pack()
        {
            rawPacket[0] = Packet.STX;
            rawPacket[Size - 1] = Packet.ETX;
            rawPacket[Size - 2] = Util.Checksum.Crc8(rawPacket, Size - 2); // size issue

            return rawPacket;
        }

        /// <summary>
        /// Clones this packet.
        /// </summary>
        /// <returns>Returns a deep copy of the current packet.</returns>
        public Object Clone()
        {
            Packet packet = new Packet();
            for(int i=0; i<Size; i++)
                packet.rawPacket[i] = rawPacket[i]; // shorter?.. CopyMemory
            return packet;
        }

        #endregion
    }
}
