/*
 * Velbus Serial Communications Library for .NET
 * Copyright (C) 2007 Velleman
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 * 
 */
#region "Imports"

using System;
using System.Collections.Generic;
using System.Text;

#endregion

namespace Velleman.Velbus
{
    /// <summary>
    /// Parser for Velbus protocol packets.
    /// </summary>
    public class PacketParser
    {
        #region "Public Fields"

        /// <summary>
        /// Internal input buffer size.
        /// </summary>
        public const int BUFFER_SIZE = 4096;

        #endregion

        /// <summary>
        /// Internal input buffer.
        /// </summary>
        private byte[] m_Buffer = new byte[BUFFER_SIZE];

        /// <summary>
        /// Current fill level of the buffer.
        /// </summary>
        private int m_Fill = 0;

        /// <summary>
        /// Required fill for the next IO operation.
        /// </summary>
        private int RequiredFill = 4;

        #region "Private Methods"

        /// <summary>
        /// Attempts to find the start of a new packet by shifting the buffer.
        /// </summary>
        private void RealignBuffer()
        {
            int i = 1;
            while ((i < m_Fill) && (m_Buffer[i] != Packet.STX))
                i += 1;

            int j = 0;
            while ((i < m_Fill))
            {
                m_Buffer[j] = m_Buffer[i];

                i += 1;
                j += 1;
            }

            m_Fill = j;
        }

        /// <summary>
        /// Analyses the start of the input buffer to see if a valid
        /// packet header is present.
        /// </summary>
        /// <returns>Returns true if a valid packet header is present in the input
        /// buffer, false otherwise.</returns>
        private bool ValidHeaderWaiting()
        {
            return ((m_Buffer[0] == Packet.STX)
                && ((m_Buffer[3] & 0x0F) <= 8)
                && ((m_Buffer[1] == 0xFB) || (m_Buffer[1] == 0xF8)));
        }

        /// <summary>
        /// Checks the input buffer to see if a valid packet body is present.
        /// </summary>
        /// <returns>Returns true if a packet body was identified, false otherwise.</returns>
        private bool ValidBodyWaiting()
        {
            int PacketSize = 6 + (m_Buffer[3] & 0x0F);
            return ((m_Buffer[PacketSize - 1] == Packet.ETX)
                && (m_Buffer[PacketSize-2] == Util.Checksum.Crc8(m_Buffer, PacketSize-3)));
        }

        /// <summary>
        /// Extracts the packets currently waiting in the input buffer.
        /// </summary>
        /// <returns>The extracted packet.</returns>
        private Packet ExtractPacket()
        {
            Packet packet = new Packet();
            packet.Priority = (m_Buffer[1] == 0xF8 ? PacketPriority.High : PacketPriority.Low);
            packet.Address = m_Buffer[2];
            packet.Rtr = ((m_Buffer[3] & 0x40) == 0x40);
            packet.DataSize = (m_Buffer[3] & 0x0F);
            for (int i = 0; i < packet.DataSize; i++)
                packet[i] = m_Buffer[4 + i];

            int k = packet.Size;
            int j = 0;
            while (k <= m_Fill)
            {
                m_Buffer[j] = m_Buffer[k];

                k += 1;
                j += 1;
            }

            return packet;
        }

        #endregion

        #region "Public Methods"

        /// <summary>
        /// Feeds data to the parser. 
        /// </summary>
        /// <param name="data">Data block to feed to the parser.</param>
        /// <param name="size">Size of the data block.</param>
        /// <exception cref="OverflowException"></exception>
        public void Feed(byte[] data, int size)
        {
            if ((data.Length + m_Fill) > PacketParser.BUFFER_SIZE)
                throw new OverflowException();
            else
            {
                for (int i = 0; i < size; i++)
                    m_Buffer[m_Fill + i] = data[i];
                m_Fill += size;
            }
        }

        /// <summary>
        /// Attempts to get the next packet from its internal buffer.
        /// </summary>
        /// <returns>Returns the next available packet, or null if no valid 
        /// packets were found.</returns>
        public Packet Next()
        {
            Packet packet = null;

            while ((m_Fill >= RequiredFill) && !ValidHeaderWaiting())
                RealignBuffer();

            if ((m_Fill >= RequiredFill) && ValidHeaderWaiting())
            {
                RequiredFill = 6 + (m_Buffer[3] & 0x0F);

                if (m_Fill >= RequiredFill)
                {
                    if (ValidBodyWaiting())
                    {
                        packet = ExtractPacket();
                        RequiredFill = 4;
                    }
                    else
                    {
                        RealignBuffer(); // force realign
                    }
                }
            }

            return packet;
        }

        #endregion
    }
}
